/* COPYRIGHT (c) 1998     Thomas Papanikolaou and Xavier Roblot
 * All rights reserved.    (papanik,roblot@math.u-bordeaux.fr)
 *
 * FILE: pari-x.y.z/src/modules/mpqs.c (thereafter called Software)
 *       (NOTE: the exact position of the file within PARI may vary).
 *
 * DESCRIPTION: Implementation of the Self-Initializing Multi-Polynomial
 * Quadratic Sieve based on code developed as part of the LiDIA project
 * (http://www.informatik.tu-darmstadt.de/TI/LiDIA/Welcome.html)
 *
 * LEGALESE: This Software is intended as a service to the PARI user
 * community, but the authors cannot be held responsible for any
 * consequences, either direct or indirect, which the use of this
 * package may have. It can be freely copied and spread for non
 * commercial purposes, as long as it is not modified and this
 * copyright notice stays attached. A free, exclusive license to
 * distribute the Software is herewith granted to the authors of PARI,
 * as stated in the PARI copyright file. However, modified versions and
 * derivatives of this Software do not become eligible for commercial
 * use, i.e. all commercial rights remain with the original authors.
 * Especially, including PARI into commercial code (either as text or
 * by linking the PARI library) does not allow the use of the Software.
 * If you want to use this Software, modified versions or derivatives of
 * this Software for commercial purposes, you have to contact the original
 * authors for a commercial license.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
/* $Id$ */
#include "pari.h"

GEN ellfacteur(GEN n, int nb_loops);
long is_odd_power(GEN x, GEN *pt, long *mask);

#ifndef SEEK_SET
#  define SEEK_SET 0
#endif

#define MPQS_STRING_LENGTH         (4 * 1024)
#define MPQS_CANDIDATE_ARRAY_SIZE  2000
#define MPQS_PRIMES_FOR_MULTIPLIER 5
#define MPQS_LP_BOUND              10000000 /* works for 32 and 64bit */

#ifdef LONG_IS_64BIT
#  define MPQS_DECIMALS_IN_LONG 19
#  define MPQS_BYTES_MASK 0x8080808080808080UL
#else
#  define MPQS_DECIMALS_IN_LONG 10
#  define MPQS_BYTES_MASK 0x80808080UL
#endif

typedef ulong *mpqs_gauss_row;
typedef long *mpqs_gauss_indices;
typedef mpqs_gauss_row *mpqs_gauss_matrix;

/* BEGIN: global variables to disappear as soon as possible */
/* names for temp. files, set in mpqs_get_filename */
static char *FREL_str;
static char *LPREL_str;
static char *LPNEW_str;
static char *LPTMP_str;

/* our own pointer to PARI's or to our own prime diffs table.
 * NB the latter is never freed, unless we need to replace it with
 * an even larger one. */
static byteptr mpqs_diffptr = NULL;
static long mpqs_prime_count = 0;

/* state flags for cleanup if we get interrupted --GN */
static char all_clean = 1;	/* set to 0 while mpqs() is busy */
#if 0				/* not used yet -- maybe wrong approach */
static char dirty_file_FREL = 0; /* set to 'r'/'w' appropriately */
static char dirty_file_LPREL = 0;
static char dirty_file_LPNEW = 0;
static char dirty_file_LPTMP = 0;
#endif
/* END: global variables to disappear as soon as possible */


/******************************/

/**
 ** determines a unique name for a file based on a short nickname
 **/

char *
pari_unique_filename(char *s);	/* in src/language/es.c */

/* name is allocated on the stack */
static char *
mpqs_get_filename(char *s)
{
  char *buf;

  s = pari_unique_filename(s);
  buf = (char*) cgeti((strlen(s) / sizeof(long)) + 2);
  return strcpy(buf, s);
}

/**
 ** opens a file s using a previously found unique name
 **/

static FILE*
mpqs_open_file(char *s, char *mode)
{
  FILE *fp = fopen(s, mode);
  if (fp == NULL)
    err(talker, "MPQS: could not open requested file %s", s);
  return fp;
}

/**
 ** closes a file fp
 **/

static void
mpqs_close_file(FILE *fp)
{
  /* This is still not right.  We should e.g. omit fflush() when
     the file was only open for reading;  we should be passed the
     pathname so we can issue a more meaningful error message... GN */
  if (fflush(fp))
    err(warner, "error whilst flushing temporary file");
  if (fclose(fp))
    err(warner, "error trying to close temporary file");
}

/**
 ** compares two `large prime' relations according to their
 ** first element (the large prime itself).
 **/

static int
mpqs_relations_cmp(const void *a, const void *b)
{
  char **sa = (char **) a;
  char **sb = (char **) b;
  long qa = atol(*sa);
  long qb = atol(*sb);
  if (qa < qb) return -1;
  else if (qa > qb) return 1;
  else return strcmp(*sa, *sb);
}

/**
 ** Given a file "filename" containing `large prime' relations, this
 ** function rearranges the file so that relations are sorted by
 ** their first elements. Works also for sorting full relations.
 ** Works in memory and overwrites the original file.
 **/

#define bummer(fn) err(talker, "error whilst writing to file %s", fn)

static long
mpqs_sort_lp_file(char *filename)
{
  FILE *TMP;
  char buf[MPQS_STRING_LENGTH];
  char **sort_table, *old_s;
  long i, j, count = 0;

  /* do we have to do this this way??? GN */
  TMP = mpqs_open_file(filename, "r");
  i = 0;
  while (fgets(buf, MPQS_STRING_LENGTH, TMP) != NULL)
    i++;

  if (i == 0)
  {
    mpqs_close_file(TMP);
    return count;
  }

  sort_table = (char **) gpmalloc((i+1) * sizeof(char *));
  for (j = 0; j < i; j++)
    sort_table[j] =
      (char *) gpmalloc(MPQS_STRING_LENGTH * sizeof(char));

  rewind(TMP);
  i = 0;
  while (fgets(sort_table[i], MPQS_STRING_LENGTH, TMP) != NULL)
    i++;
  mpqs_close_file(TMP);

  qsort(sort_table, i, sizeof(char *), mpqs_relations_cmp);

  TMP = mpqs_open_file(filename, "w");
  old_s = sort_table[0];
  if (fputs(sort_table[0], TMP) < 0)
    bummer(filename);
  count = 1;
  for(j = 1; j < i; j++)
  {
    if (strcmp(old_s, sort_table[j]) != 0)
    {
      if (fputs(sort_table[j], TMP) < 0)
	bummer(filename);
      count++;
    }
    old_s = sort_table[j];
  }
  mpqs_close_file(TMP);

  for (j = 0; j < i; j++)
    free(sort_table[j]);
  free(sort_table);

  return count;
}

/**
 ** appends contents of file fp1 to file fp of name filename
 ** (auxiliary routine for merge sort).
 ** fp assumed open for appending and fp1 for reading.
 **/

static void
mpqs_append_file(FILE *fp, FILE *fp1, char *filename)
{
  char line[MPQS_STRING_LENGTH];
  while (fgets(line, MPQS_STRING_LENGTH, fp1) != NULL)
    if (fputs(line, fp) < 0)
      err(talker, "error whilst appending to file %s", filename);
  if (fflush(fp))
    err(warner, "error whilst flushing file %s", filename);
  if (fclose(fp))
    err(warner, "error trying to close file %s", filename);
}

/**
 ** Does a merge-sort on the files lprel and lpnew;
 ** assumes that lp and lpnew are already sorted.
 ** Creates/truncates the TMP file and leaves it open.
 **/

static long
mpqs_mergesort_lp_file(FILE *LPREL, FILE *LPNEW)
{
  FILE *TMP = mpqs_open_file(LPTMP_str, "w");
  char line_new[MPQS_STRING_LENGTH], line[MPQS_STRING_LENGTH];
  long q_new, q, i = 0, c = 0;

  /* if LPNEW is empty, copy the remainder of LPREL to TMP */

  if (fgets(line_new, MPQS_STRING_LENGTH, LPNEW) == NULL)
  {
    mpqs_append_file(TMP, LPREL, LPTMP_str);
    return 0;
  }

  /* and vice versa */

  if (fgets(line, MPQS_STRING_LENGTH, LPREL) == NULL)
  {
    if (fputs(line_new, TMP) < 0)
      bummer(LPTMP_str);
    mpqs_append_file(TMP, LPNEW, LPTMP_str);
    return 0;
  }

  /* both LPNEW and LPREL are not empty */
  q_new = atol(line_new);
  q = atol(line);

  for(;;)
  {
    while (q < q_new)
    {
      if (fputs(line, TMP) < 0)
	bummer(LPTMP_str);
      if (fgets(line, MPQS_STRING_LENGTH, LPREL) == NULL)
      {
	if (fputs(line_new, TMP) < 0)
	  bummer(LPTMP_str);
	mpqs_append_file(TMP, LPNEW, LPTMP_str);
	return c;
      }
      else
	q = atol(line);
    }

    /* q >= q_new */

    while (q > q_new)
    {
      if (fputs(line_new, TMP) < 0)
	bummer(LPTMP_str);
      if (fgets(line_new, MPQS_STRING_LENGTH, LPNEW) == NULL)
      {
	if (fputs(line, TMP) < 0)
	  bummer(LPTMP_str);
	mpqs_append_file(TMP, LPREL, LPTMP_str);
	return c;
      }
      else
	q_new = atol(line_new);
    }

    /* q <= q_new */

    i = 0;
    while (q == q_new)
    {
      if (strcmp(line_new, line) == 0)
      {
	if (fputs(line, TMP) < 0)
	  bummer(LPTMP_str);

	if (fgets(line, MPQS_STRING_LENGTH, LPREL) == NULL)
	{
	  mpqs_append_file(TMP, LPNEW, LPTMP_str);
	  return c + i;
	}
	else
	  q = atol(line);

	if (fgets(line_new, MPQS_STRING_LENGTH, LPNEW) == NULL)
	{
	  if (fputs(line, TMP) < 0)
	    bummer(LPTMP_str);
	  mpqs_append_file(TMP, LPREL, LPTMP_str);
	  return c + i;
	}
	else
	  q_new = atol(line_new);
      }
      else
      {
	i++;
	if (fputs(line, TMP) < 0)
	  bummer(LPTMP_str);

	if (fgets(line, MPQS_STRING_LENGTH, LPREL) == NULL)
	{
	  mpqs_append_file(TMP, LPNEW, LPTMP_str);
	  return c + i;
	}
	else
	  q = atol(line);
      }
    }
    c += i;
  }
}

/******************************/
/**
 ** return next prime larger than p, using *primes_ptr on the
 ** diffptr table first and pari's other wits after that
 **/

static byteptr
mpqs_iterate_primes(long *p, byteptr primes_ptr)
{
  long prime = *p;
  if (*primes_ptr)
    prime += *primes_ptr++;
  else
  {
    long av = avma;
    prime = itos(nextprime(stoi(prime + 1)));
    avma = av;
  }
  *p = prime;
  return primes_ptr;
}


/**
 ** return the multiplier k for N to use in MPQS
 **/

static long
mpqs_find_k(GEN N, long tries)
{

  static long cand_table[] = { 1,3,5,7,11,13,15,17,19,21,23,
                               29,31,33,35,37,39,41,43,47,51,53,
                               55,57,59,61,65,67,69,71,73,79,83,
                               85,89,91,93,95,97,101,103,105,107,
                               109,113,127,131,137,139};
  long av = avma;
  long best_k = 1, k, p, N_mod_4 = smodis(N, 4), x;
  GEN kN;
  double best_value = 1, value, dp;
  long i, j;
  byteptr primes_ptr;

  for (i = 0; i < MPQS_PRIMES_FOR_MULTIPLIER; i++)
  {
    k = cand_table[i];
    x = (k * N_mod_4) % 4;
    if (x == 1)
    {
      value = -0.7 * log2 ((double) k);
      kN = mulis(N, k);
      if (smodis(kN, 8) == 1)
	value += 1.38629;

      j = 0; p = 0;
      primes_ptr = mpqs_diffptr;
      while (j <= tries)
      {
	primes_ptr = mpqs_iterate_primes(&p, primes_ptr);
	if (kross(smodis(kN, p), p) == 1)
        {
	  j++;
	  dp = log2((double) p) / p;
	  value += (k % p == 0) ? dp : 2 * dp;
	}
      }

      if (value > best_value)
      {
	best_value = value;
	best_k = k;
      }
    }
  }

  avma = av;
  return best_k;
}

/******************************/
/**
 ** guesstimate up to what size we're going to need precomputed
 ** small primes
 **/

static long
mpqs_find_maxprime(long size)
{
  double x;

  if (size < 16000)
    return 176000;

  x  = log((double)size);
  x += log(x);
  x -= 0.9427;
  x *= size;

  return (long)x;
}

/**
 ** return the number of primes initialized in PARI
 **/

static long
mpqs_count_primes()
{
  byteptr p = mpqs_diffptr;

  for ( ; *p; p++) /* empty */;
  return (p - mpqs_diffptr);
}

/**
 ** create a factor base of size primes p_i with the
 ** property that legendre(k*N, p_i) == 0, 1
 ** FB[0] contains the size of the factor base
 ** FB[1] contains -1 which allows to handle negative values of Q(x).
 ** If a prime factor of N is found during the construction, it
 ** is returned in f, otherwise f = 0  (such a factor will necessarily
 ** fit into a C long).
 **/

static long*
mpqs_create_FB(long size, GEN kN, long k, long *f)
{
  long i, kr, p = 0;
  long *FB;
  byteptr primes_ptr;

  FB = (long *) calloc(size + 3, sizeof(long));
  if (!FB) err(memer);
  FB[0] = size;
  FB[1] = -1;

  if ((mpqs_prime_count ? mpqs_prime_count : mpqs_count_primes())
      < 3 * size)
  {
    long newsize = 3 * mpqs_find_maxprime(size);
    if (mpqs_diffptr != diffptr) free((void *) mpqs_diffptr);
    if (DEBUGLEVEL >= 2)
    {
      fprintferr("MPQS: precomputing auxiliary primes up to %ld\n",
		 newsize);
    }
    mpqs_diffptr = initprimes(newsize);
    /* count 'em once and remember the result */
    mpqs_prime_count = mpqs_count_primes();
  }

  primes_ptr = mpqs_diffptr;
  for (i = 2; i < size + 2; )
  {
    primes_ptr = mpqs_iterate_primes(&p, primes_ptr);

    if (p != k)
    {
      kr = kross(smodis(kN, p), p);
      if (kr != -1)
      {
	if (kr == 0)
	{
	  *f = p;
	  return FB;
	}
	FB[i] = p, i++;
      }
    }
  }

  FB[i] = 0;
  if (DEBUGLEVEL >= 6)
    fprintferr("MPQS: last available index in FB is %ld\n", i - 1);

  *f = 0;
  return FB;
}

/******************************/

/**
 ** see below for the format of the parameter table.
 ** These values are somewhat experimental and can be improved by
 ** further extensive testing.
 **/

static double
mpqs_parameters[66][7] =
{
  {15, 0.8, 1200, 35, 3, 10, 2},
  {16, 0.8, 1400, 35, 3, 10, 2},
  {17, 0.8, 3000, 40, 3, 10, 2},
  {18, 0.8, 3000, 60, 3, 10, 2},
  {19, 0.8, 3600, 80, 3, 10, 2},
  {20, 0.8, 4000, 100, 3, 10, 2},
  {21, 0.8, 4300, 100, 3, 10, 2},
  {22, 0.8, 4500, 120, 3, 10, 3},
  {23, 0.8, 4800, 140, 3, 10, 3},
  {24, 0.8, 5000, 160, 3, 12, 4},
  {25, 0.8, 5000, 180, 3, 12, 4},
  {26, 0.9, 6000, 200, 5, 8, 4},
  {27, 1.17, 6000, 220, 5, 8, 5},
  {28, 1.17, 6500, 240, 5, 8, 5},
  {29, 1.17, 6500, 260, 5, 8, 5},
  {30, 1.36, 7000, 325, 5, 8, 5},
  {31, 1.36, 7000, 355, 5, 8, 5},
  {32, 1.36, 7500, 375, 5, 8, 5},
  {33, 1.43, 7500, 400, 6, 8, 6},
  {34, 1.43, 7500, 425, 6, 8, 6},
  {35, 1.43, 7500, 550, 6, 10, 6},
  {36, 1.43, 8000, 650, 6, 10, 6},
  {37, 1.69, 9000, 750, 6, 10, 7},
  {38, 1.69, 10000, 850, 6, 10, 7},
  {39, 1.69, 11000, 950, 6, 10, 7},
  {40, 1.69, 14000, 1000, 6, 10, 7},
  {41, 1.69, 14000, 1150, 6, 10, 8},
  {42, 1.69, 15000, 1300, 6, 10, 8},
  {43, 1.69, 15000, 1600, 6, 10, 8},
  {44, 1.69, 15000, 1900, 7, 10, 9},
  {45, 1.69, 15000, 2200, 7, 10, 9},
  {46, 1.69, 20000, 2500, 7, 10, 9},
  {47, 1.69, 25000, 2500, 7, 10, 10},
  {48, 1.69, 27500, 2700, 7, 10, 10},
  {49, 1.69, 30000, 2800, 7, 10, 10},
  {50, 1.75, 35000, 2900, 7, 11, 10},
  {51, 1.75, 40000, 3000, 7, 11, 10},
  {52, 1.85, 50000, 3200, 7, 11, 11},
  {53, 1.85, 50000, 3500, 7, 11, 11},
  {54, 1.95, 80000, 3800, 7, 12, 11},
  {55, 1.95, 90000, 4100, 7, 12, 11},
  {56, 1.95, 100000, 4400, 7, 12, 11},
  {57, 1.95, 80000, 4700, 8, 12, 12},
  {58, 1.95, 80000, 5000, 8, 12, 12},
  {59, 2.15, 130000, 5500, 8, 12, 12},
  {60, 2.15, 140000, 5800, 8, 12, 12},
  {61, 2.15, 150000, 6100, 8, 13, 13},
  {62, 2.15, 160000, 6400, 8, 13, 13},
  {63, 2.35, 170000, 6700, 8, 13, 13},
  {64, 2.35, 180000, 7000, 8, 13, 13},
  {65, 2.35, 190000, 7300, 8, 13, 13},
  {66, 2.35, 200000, 7600, 8, 13, 13},
  {67, 2.4, 150000, 7900, 8, 13, 13},
  {68, 2.4, 150000, 8200, 8, 14, 13},
  {69, 2.4, 130000, 8600, 8, 14, 13},
  {70, 2.45, 130000, 8800, 8, 14, 13},
  {71, 2.45, 130000, 8800, 9, 14, 13},
  {72, 2.4, 260000, 9400, 9, 14, 13},
  {73, 2.4, 270000, 9700, 9, 14, 13},
  {74, 2.4, 280000, 9000, 9, 14, 13},
  {75, 2.6, 140000, 9000, 9, 14, 13},
  {76, 2.6, 160000, 9400, 9, 14, 13},
  {77, 2.6, 180000, 9600, 9, 14, 13},
  {78, 2.6, 200000, 9800, 9, 14, 13},
  {79, 2.65, 220000, 10000, 9, 14, 13},
  {80, 2.65, 250000, 10500, 9, 14, 13}};

/**
 ** Return the appropriate parameters for the initialization of MPQS
 **
 ** the parameter table for mpqs has the following format:
 **
 ** d    -- decimal digits
 ** e    -- approximation accuracy
 ** M    -- size sieving interval
 ** s    -- size FB,
 ** fA   -- # prime factors in A
 ** pA   -- total # primes for determination of A,
 ** [maxB -- how many Bs to use before choosing a new A -- determined by fA]
 ** si   -- initial sieving index
 **/

static void
mpqs_read_parameters(ulong d, double *e, ulong *M, ulong *s, ulong *fA,
                     ulong *pA, long *maxB, ulong *si)
{
  long i = d;
  if (i < 15) i = 15;
  if (i > 80) i = 80;
  i -= 15;

  *e = mpqs_parameters[i][1];
  *M = (ulong)mpqs_parameters[i][2];
  *s = (ulong)mpqs_parameters[i][3];
  *fA = (ulong)mpqs_parameters[i][4];
  *pA = (ulong)mpqs_parameters[i][5];
  *maxB = (long) pow(2.0, mpqs_parameters[i][4] - 1.0);
  *si = (ulong)mpqs_parameters[i][6];
}

/******************************/

#if 0
/**
 ** count the number of 1s in the binary representation of x
 **/
static ulong
mpqs_count_ones(ulong x)
{
  ulong s = 0;
  while (x) s += (x & 1UL), x >>= 1;
  return s;
}
#else
/**
 ** increment *x (!=0) to the smallest larger value which has the same
 ** number of 1s in its binary representation.  Will only be used in the
 ** difficult cases, so we can take our time.  Cases which wrap around
 ** can be detected by the caller afterwards as long as we keep
 ** total_no_of_primes_for_A strictly less than BITS_IN_LONG.
 **/
static void
mpqs_increment(ulong *x)
{
  ulong r1_mask, r01_mask, slider=1;

  /* isolate rightmost 1 */
  r1_mask = ((*x ^ (*x - 1)) + 1) >> 1;
  /* isolate rightmost 1 which has a 0 to its left */
  r01_mask = ((*x ^ (*x + r1_mask)) + r1_mask) >> 2;
  /* simple cases: */
  if (r1_mask == r01_mask) { *x += r1_mask; return; }
  if (r1_mask == 1) { *x += r01_mask; return; }
  /* general case -- idea:  add r01_mask, kill off as many 1 bits as possible
     to its right while at the same time filling in 1 bits from the LSB. */
  if (r1_mask == 2) { *x += (r01_mask>>1) + 1; return; }
  while (r01_mask > r1_mask && slider < r1_mask)
  {
    r01_mask >>= 1; slider <<= 1;
  }
  *x += r01_mask + slider - 1;
}
#endif

static long
mpqs_invsmod(long b, long p)
{
  long v1 = 0, v2 = 1, v3, r, oldp = p;

  while (b > 1)
  {
    v3 = v1 - (p / b) * v2; v1 = v2; v2 = v3;
    r = p % b; p = b; b = r;
  }

  if (v2 < 0) v2 += oldp;
  return v2;
}

/**
 ** compute coefficients of the sieving polynomial for self initializing
 ** variant. Coefficients A and B are returned and several tables are
 ** updated.  See Thomas Sosnowski's diploma thesis.
 **/

static void
mpqs_self_init(GEN A, GEN B, GEN kN, long *FB, long *sqrt_mod_p_kN,
               long *start_1, long *start_2, ulong no_of_primes_in_A,
               ulong total_no_of_primes_for_A, ulong M, long **inv_A_H_i,
               long *Q_prime, long *Q_prime_glob, GEN* H_i, long index_i,
               long start_index_FB_for_A, long *inv_A2, GEN inv_A4,
               ulong *bin_index)
{
  long av;
  GEN p1, p2;
  ulong size_of_FB, nu_2;
  long i, j, p, M_mod_p, tmp, tmp1, tmp2;

  if (index_i == 0)
  { /* compute first polynomial with new A */
#if 0
    (*bin_index)++;
    while (mpqs_count_ones(*bin_index) != no_of_primes_in_A)
      (*bin_index)++;
#else
    /* ultra-fast 32-way computed jump handles 26 out of 32 cases */
    if (!*bin_index) *bin_index = (1ul << no_of_primes_in_A) - 1;
    else
    {
      switch (*bin_index & 0x1f)
      {
      case 1: case 5: case 9: case 13: case 17: case 21: case 25: case 29:
	(*bin_index)++; break;
      case 2: case 10: case 18: case 26:
      case 3: case 11: case 19: case 27:
	(*bin_index) += 2; break;
      case 6: case 22:
	(*bin_index) += 3; break;
      case 4: case 20:
      case 7: case 23:
	(*bin_index) += 4; break;
      case 12: case 14:
	(*bin_index) += 5; break;
      case 8: case 15:
	(*bin_index) += 8; break;
      default:
	mpqs_increment(bin_index); break;
      }
      if (*bin_index >= (1ul << total_no_of_primes_for_A)) /* wraparound */
      {
	if (DEBUGLEVEL >= 2)
	  fprintf(stderr, "MPQS: bin_index wraparound\n");
	/* complain:  caller should increase some parameters... */
	return;
	/* caller should repeat the above test; if we simply increase the
	   total_no_of_primes_for_A, then bin_index now contains something
	   suitable for carrying on -- we'll probably end up not using
	   that particular index and skipping ahead to the next one, but
	   this isn't a problem --GN */
      }
    }
#endif
    /* determine primes used for A in this iteration */
    for (i = 0, j = 0; j < total_no_of_primes_for_A; j++)
    {
      if (*bin_index & (1 << j))
      {
	Q_prime[i] = Q_prime_glob[j];
	i++;
      }
    }

    /*
     * compute coefficient A using
     * A = prod (Q_prime[i], 0 < i < no_of_primes_in_A)
     */

    av = avma;
    p1 = stoi(Q_prime[0]);
    for (i = 1; i < no_of_primes_in_A; i++)
      p1 = mulis(p1, Q_prime[i]);
    affii(p1, A);
    avma = av;

    /*
     * let bound := no_of_primes_in_A. Compute H_i, 0 <= i < bound.
     */

    for (i = 0; i < no_of_primes_in_A; i++)
    {
      av = avma;
      p = Q_prime[i];
      p1 = divis(A, p);
      tmp = smodis(p1, p);
      p1 = mulis(p1, mpqs_invsmod(tmp, p));
      tmp = itos(mpsqrtmod(modis(kN, p), stoi(p)));
      p1 = mulis(p1, tmp);
      affii(modii(p1, A), H_i[i]);
      avma = av;
    }

    /*
     * Each B is of the form sum(v_i * H_i, 0 <= i < bound)
     * where each v_j is +1 or -1. This has to be done only once
     * (for index_i==0) for the coefficient A; if index_i > 0 then
     * there is a linear recursion for B
     */

    /* compute actual B coefficient, by taking all v_i == 1 */
    av = avma;
    p1 = H_i[0];
    for (i = 1; i < no_of_primes_in_A; i++)
      p1 = addii(p1, H_i[i]);
    affii(p1, B);
    avma = av;

    /* ensure B = 1 mod 4 */
    if (mod2(B) == 0)
    {
      /* B += (A % 4) * A; */
      av = avma;
      p1 = addii(B, mulsi(mod4(A), A));
      affii(p1, B);
      avma = av;
    }

    /* compute inv_A2[i] = 1/(2*A) mod p_i */
    av = avma;
    p1 = shifti(A, 1);
    size_of_FB = FB[0] + 1;
    for (i = 2; i <= size_of_FB; i++)
      inv_A2[i] = mpqs_invsmod(smodis(p1, FB[i]), FB[i]);
    avma = av;

    /* compute inv_A_H_i[i][j] = 1/A * H_i[i] mod p_j */
    for (i = 0; i < no_of_primes_in_A; i++)
    {
      for (j = 2; j <= size_of_FB; j++)
      {
	av = avma;
	p1 = mulis(H_i[i], inv_A2[j] << 1);
	inv_A_H_i[i][j] = smodis(p1, FB[j]);
	avma = av;
      }
    }

    /* compute the roots z1, z2, of the polynomial Q(x) mod p_j and
     * initialize start_1[i] with the first value p_i | Q(z1 + i p_j)
     * initialize start_2[i] with the first value p_i | Q(z2 + i p_j)
     */
    for (j = 2; j <= size_of_FB; j++)
    {
      p = FB[j];
      M_mod_p = M % p;

      tmp = p - smodis(B, p); /* tmp = -B mod p */

      tmp1 = (tmp - sqrt_mod_p_kN[j]) % p;
      if (tmp1 < 0) tmp1 += p;
      tmp1 = mulssmod(tmp1, inv_A2[j], p);
      tmp1 = (tmp1 + M_mod_p) % p;

      tmp2 = (tmp + sqrt_mod_p_kN[j]) % p;
      if (tmp2 < 0) tmp2 += p;
      tmp2 = mulssmod(tmp2, inv_A2[j], p);
      tmp2  = (tmp2 + M_mod_p) % p;

      start_1[j] = (tmp1 < 0) ? tmp1 + p : tmp1;
      start_2[j] = (tmp2 < 0) ? tmp2 + p : tmp2;
    }

    /* determine 1/(4A) mod kN */
    av = avma;
    p1 = shifti(A, 2);
    p1 = mpinvmod(p1, kN);
    affii(p1, inv_A4);
    avma = av;
  }
  else
  {
    /* no "real" computation -- use recursive formula
     * first: update of B, compute B[index_i], index_i > 0
     */

    /* compute nu_2 = nu_2(index_i) */
    nu_2 = 0;
    j = index_i;
    while ((j & 1) == 0)
    {
      nu_2++;
      j >>= 1;
    }

    av = avma;
    p1 = shifti(H_i[nu_2], 1);

    if ((((j+1)/2) & 1) == 1)
    {
      i = -1;
      p1 = subii(B, p1);
    }
    else
    {
      i = 1;
      p1 = addii(B, p1);
    }
    affii(p1, B);
    avma = av;

    /* determine new starting positions for sieving */
    size_of_FB = FB[0] + 1;

    if (i == -1)
    {
      for (j = 2; j <= size_of_FB; j++)
      {
	p = FB[j];
	start_1[j] += inv_A_H_i[nu_2][j];
	if (start_1[j] >= p)
	  start_1[j] -= p;
	start_2[j] += inv_A_H_i[nu_2][j];
	if (start_2[j] >= p)
	  start_2[j] -= p;
      }
    }
    else
    {
      for (j = 2; j <= size_of_FB; j++)
      {
	p = FB[j];
	start_1[j] -= inv_A_H_i[nu_2][j];
	if (start_1[j] < 0)
	  start_1[j] += p;
	start_2[j] -= inv_A_H_i[nu_2][j];
	if (start_2[j] < 0)
	  start_2[j] += p;
      }
    }
  }

  /* p=2 is a special case */
  if (FB[2] == 2)
  {
    start_1[2] = 0; start_2[2] = 1;
  }


  /* now compute zeros of polynomials that have only one zero mod p
     because p divides the coefficient A */

  /* compute coefficient -C */
  av = avma;
  p1 = subii(kN, sqri(B));
  p2 = divii(p1, shifti(A, 2));

  for (j = 1; j <= total_no_of_primes_for_A; j++)
    if (*bin_index & (1 << (j-1)))
    {
      p = FB[start_index_FB_for_A + j];
      tmp = mpqs_invsmod(smodis(B, p), p);
      tmp2 = smodis (p2, p);
      tmp = mulssmod(tmp2, tmp, p);
      start_1[start_index_FB_for_A + j] =
	start_2[start_index_FB_for_A + j] = (tmp + M) % p;
    }
  avma = av;

#ifdef MPQS_DEBUG
  {
    /*
     * check correctness of the roots z1, z2 mod p_i by evaluting
     * A * z1^2 + B * z1 + C mod p_i and checking against 0 mod p_i
     * C is written as (B*B-kN)/(4*A)
     */
    size_of_FB = FB[0] + 1;
    for (j = 3; j <= size_of_FB; j++)
    {
      av = avma;
      p = FB[j];
      M_mod_p = M % p;

      p1 = mulis(A, start_1[j] - M_mod_p);
      p1 = addii(p1, B);
      p1 = mulis(p1, start_1[j] - M_mod_p);
      p2 = divii(subii(sqri(B), kN), shifti(A, 2));
      p1 = addii(p1, p2);
      if (smodis(p1, p) != 0)
      {
	fprintferr("MPQS: (1) j = %ld p = %ld\n", j, p);
	fprintferr("MPQS: (1) index_i = %ld\n", index_i);
	fprintferr("MPQS: A = %Z\n", (long)A);
	fprintferr("MPQS: B = %Z\n", (long)B);
	fprintferr("MPQS: C = %Z\n", (long)p1);
	fprintferr("MPQS: z1 = %ld\n", start_1[j] - M_mod_p);
	err(talker, "MPQS: self_init: found wrong polynomial in (1)");
      }

      p1 = mulis(A, start_2[j]-M_mod_p);
      p1 = addii(p1, B);
      p1 = mulis(p1, start_2[j]-M_mod_p);
      p2 = divii(subii(sqri(B), kN), shifti(A, 2));
      p1 = addii(p1, p2);
      if (smodis(p1, p) != 0)
      {
	fprintferr("MPQS: (2) j = %ld p = %ld\n", j, p);
	fprintferr("MPQS: (2) index_i = %ld\n", index_i);
	fprintferr("MPQS: A = %Z\n", (long)A);
	fprintferr("MPQS: B = %Z\n" (long)B);
	fprintferr("MPQS: C = %Z\n" (long)p1);
	fprintferr("MPQS: z2 = %ld\n", start_2[j] - M_mod_p);
	err(talker, "MPQS: self_init: found wrong polynomial in (2)");
      }
      avma = av;
    }
    if (DEBUGLEVEL >= 4)
      fprintferr("MPQS: checking of roots of Q(x) was successful\n");
  }
#endif

}

/******************************/

/**
 **  Main sieving routine:
 **
 **  FB     is a pointer to an array which holds the factor base
 **  log_FB is a pointer to an array which holds the approximations for
 **         the logarithms of the factor base elements
 **  start_1, start_2
 **         are arrays for starting points for different FB elements
 **  sieve_array
 **         points to a sieve array
 **  sieve_array_end
 **         points to the end of the sieve array
 **  M      is the size of the sieving interval
 **  starting_sieving_index
 **         marks the first FB element which is used for sieving
 **
 **/

#ifdef INLINE
 INLINE
#endif
void
mpqs_sieve_p(unsigned char *begin, unsigned char *end,
             long p_times_4, long p, unsigned char log_p)
{
  register unsigned char *e = end - p_times_4;
  while (begin <= e)
  {
    (*begin) += log_p, begin += p;
    (*begin) += log_p, begin += p;
    (*begin) += log_p, begin += p;
    (*begin) += log_p, begin += p;
  }
  while (begin <= end)
    (*begin) += log_p, begin += p;
}

static void
mpqs_sieve(long *FB, unsigned char *log_FB, long *start_1, long *start_2,
	   unsigned char *sieve_array, unsigned char *sieve_array_end,
	   ulong M, ulong starting_sieving_index)
{
  register long p_times_4, p, *ptr_FB, start1, start2, l;
  register unsigned char log_p;

  memset (sieve_array, 0, (M << 1) * sizeof (unsigned char));

  l = starting_sieving_index;
  ptr_FB = &FB[l];

  while ((p = *ptr_FB++) != 0)
  {
    p_times_4 = p * 4;
    log_p = log_FB[l];
    start1 = start_1[l];
    start2 = start_2[l];

    /* sieve with FB[l] from start_1[l] */
    /* if start1 != start2 sieve with FB[l] from start_2[l] */
    if (start1 == start2)
      mpqs_sieve_p(sieve_array + start1, sieve_array_end,
		   p_times_4, p, log_p);
    else
    {
      mpqs_sieve_p(sieve_array + start1, sieve_array_end,
		   p_times_4, p, log_p);
      mpqs_sieve_p(sieve_array + start2, sieve_array_end,
		   p_times_4, p, log_p);
    }

    l++;
  }
}

/******************************/

static long
mpqs_eval_sieve(unsigned char *sieve_array, ulong M, long *candidates)
{
  ulong M_2 = M << 1;
  int i = 0, count, q, r;
  unsigned char *ptr = sieve_array;

  r = M_2 % 4;
  q = M_2  - r;
  count = 0;
  for (i = 0; i < q; i += 4, ptr += 4)
  {
    if (*(ptr+0) > 128) { candidates[count] = i+0; count++; }
    if (*(ptr+1) > 128) { candidates[count] = i+1; count++; }
    if (*(ptr+2) > 128) { candidates[count] = i+2; count++; }
    if (*(ptr+3) > 128) { candidates[count] = i+3; count++; }
  }
  switch (r)
  {
    case 3:
      if (*(ptr+0) > 128) { candidates[count] = i+0; count++; }
      if (*(ptr+1) > 128) { candidates[count] = i+1; count++; }
      if (*(ptr+2) > 128) { candidates[count] = i+2; count++; }
      break;
    case 2:
      if (*(ptr+0) > 128) { candidates[count] = i+0; count++; }
      if (*(ptr+1) > 128) { candidates[count] = i+1; count++; }
      break;
    case 1:
      if (*(ptr+0) > 128) { candidates[count] = i+0; count++; }
      break;
    default:
      break;
  }
  candidates[count] = 0;
  return count;
}

/**
 **  Main relation routine:
 **
 **  FB     is a pointer to an array which holds the factor base
 **  log_FB is a pointer to an array which holds the approximations for
 **         the logarithms of the factor base elements
 **  start_1, start_2
 **         are arrays for starting points for different FB elements
 **  M      is the size of the sieving interval
 **
 **/

#ifdef INLINE
INLINE void
mpqs_add_factor(char *s, ulong ei, ulong pi)
{
  sprintf(s + strlen(s), " %lu %lu", ei, pi);
}
#else
#define mpqs_add_factor(s,ei,pi) \
	sprintf(s + strlen(s), " %lu %lu", (ulong)ei, (ulong)pi)
#endif

/**
 ** divide and return the rest
 **/

static ulong
mpqs_div_rem(GEN x, long y, GEN *q)
{
  *q = divis(x, y);
  return hiremainder;
}

static long
mpqs_eval_candidates(GEN A, GEN inv_A4, GEN B, GEN kN, long k,
		     double sqrt_kN,
		     long *FB, long *start_1, long *start_2,
		     ulong no_of_primes_in_A,
		     ulong total_no_of_primes_for_A,
		     ulong M, long *Q_prime, long *candidates,
		     long number_of_candidates,
		     long start_index_FB_for_A,
		     FILE *FREL, FILE *LPREL)
{
  GEN Qx, A_2x_plus_B, Qx_div_p;
  double a, b, inv_2_a;
  long av, powers_of_2, p, tmp_p, rest_p;
  long z1, z2, x, number_of_relations, x_minus_M;
  long max_index_FB_for_A;
  char *relations;
  ulong i, pi, ei;
  int pi_A;
  int p_divides_A;

  /* FIXME: this is probably right !!! */

  /*
   * compute the roots of Q(X); this will be used to find the sign of
   * Q(x) after reducing mod kN
   */

  a = gtodouble(A);
  inv_2_a = 1 / (2.0 * a);
  b = gtodouble(B);
  z1 = (long) ((-b - sqrt_kN) * inv_2_a);
  z2 = (long) ((-b + sqrt_kN) * inv_2_a);

  number_of_relations = 0;
  /* FIXME: this is definitely too loooooong ... */
  relations =
    (char *) gpmalloc(1000 * MPQS_DECIMALS_IN_LONG * sizeof(char));
  relations[0] = '\0';

  i = 0;
  while (i < number_of_candidates)
  {
    x = candidates[i++];
    av = avma;

    /* A_2x_plus_B = (A*(2x)+B), Qx = (A*(2x)+B)^2/(4*A) = Q(x) */
    x_minus_M = x - M;
    A_2x_plus_B = modii(addii(mulis(A, x_minus_M << 1), B), kN);
    Qx = modii(sqri(A_2x_plus_B), kN);

#ifdef MPQS_DEBUG
    {
      long av = avma;
      GEN g = mppgcd(Qx, kN);
      if (kronecker(divii(Qx, g), divii(kN, g)) != 1)
	err(talker, "MPQS: 4*A*Q(x) is not a square (mod kN)");
      avma = av;
    }
#endif

    Qx = modii(mulii(Qx, inv_A4), kN);

    /* check the sign of Qx */
    if (z1 < x_minus_M && x_minus_M < z2)
    {
       Qx = subii(kN, Qx);
       /* i = 1, ei = 1, pi */
       mpqs_add_factor(relations, 1, 1);
    }

    /* divide by powers of 2; note that 4*A*Q(x) we add 2^2 */
    powers_of_2 = vali(Qx);
    Qx = shifti(Qx, -powers_of_2);
    mpqs_add_factor(relations, powers_of_2 + 2, 2);

    /* we handled the case p = 2 already */
    pi = 3;

    /* FB[pi] .. FB[start_index_FB_for_A - 1] do not divide A */
    while (pi <= start_index_FB_for_A)
    {
      p = FB[pi];
      tmp_p = x % p;

      if (tmp_p == start_1[pi] || tmp_p == start_2[pi])
      {
        /* p divides Q(x) */
        rest_p = mpqs_div_rem(Qx, p, &Qx_div_p);
        ei = 1;

        while (rest_p == 0)
        {
	  Qx = Qx_div_p;
	  rest_p = mpqs_div_rem(Qx, p, &Qx_div_p);
	  if (rest_p == 0)
	    ei++;
        }
        mpqs_add_factor(relations, ei, pi);

      }
      pi++;
    }

    /* p = FB[start_index_FB_for_A] ... FB[max_index_FB_for_A] divides A
       if p is in Q_prime[]; remember A = prod(Q_prime[i]) */

    max_index_FB_for_A = start_index_FB_for_A + total_no_of_primes_for_A;
    while (pi <= max_index_FB_for_A)
    {
      p = FB[pi];
      tmp_p = x % p;

      pi_A = 0;
      p_divides_A = 0;
      while (p_divides_A == 0 && pi_A < no_of_primes_in_A)
      {
	if (p == Q_prime[pi_A])
	  p_divides_A = 1;
	pi_A++;
      }

      ei = p_divides_A;
      if (tmp_p == start_1[pi] || tmp_p == start_2[pi])
      {
        /* p divides Q(x) and probably A */
        rest_p = mpqs_div_rem(Qx, p, &Qx_div_p);
        ei++;

        while (rest_p == 0)
        {
	  Qx = Qx_div_p;
	  rest_p = mpqs_div_rem(Qx, p, &Qx_div_p);
	  if (rest_p == 0)
	    ei++;
        }
      }
      if (ei)
        mpqs_add_factor(relations, ei, pi);

      pi++;
    }

    /* FB[max_index_FB_for_A + 1] ... FB[end] */
    while ((p = FB[pi]) != 0)
    {
      tmp_p = x % p;

      if (tmp_p == start_1[pi] || tmp_p == start_2[pi])
      {
        /* p divides Q(x) */
        rest_p = mpqs_div_rem(Qx, p, &Qx_div_p);
        ei = 1;

        while (rest_p == 0)
        {
	  Qx = Qx_div_p;
	  rest_p = mpqs_div_rem(Qx, p, &Qx_div_p);
	  if (rest_p == 0)
	    ei++;
        }
        mpqs_add_factor(relations, ei, pi);
      }

      pi++;
    }

    if (gcmp1(Qx))
    {
      char *Qxstring = (char *) gen2str(A_2x_plus_B);
      fprintf(FREL, "%s :%s 0\n", Qxstring, relations);
      number_of_relations++;

#ifdef MPQS_DEBUG
      {
	GEN Qx_2, prod_pi_ei, pi_ei;
	char *s;
	long pi, ei, av;

	av = avma;
	Qx_2 = modii(sqri(A_2x_plus_B), kN);
	s = strtok(relations, " \n");
	prod_pi_ei = gun;
	while (s != NULL)
        {
	  ei = atol(s);
	  s = strtok(NULL, " \n");
	  pi = atol(s);
	  pi_ei = powmodulo(stoi(FB[pi]), stoi(ei), kN);
	  prod_pi_ei = modii(mulii(prod_pi_ei, pi_ei), kN);
	  s = strtok(NULL, " \n");
	}
	avma = av;

	if (gcmp(Qx_2, prod_pi_ei) != 0)
	  err(talker, "MPQS: wrong full relation found!!");
      }
#endif

      free(Qxstring);
    }
    else if (cmpis(Qx, MPQS_LP_BOUND) <= 0)
    {
      /* if (mpqs_isprime(itos(Qx))) */
      if (itos(Qx) % k != 0)
      {
	char *Qxstring = (char *) gen2str(A_2x_plus_B);
	char *L1string = (char *) gen2str(Qx);
	strcat(relations, " 0");
	fprintf(LPREL, "%s @ %s :%s\n", L1string, Qxstring, relations);

#ifdef MPQS_DEBUG
	{
          GEN Qx_2, prod_pi_ei, pi_ei;
          char *s;
          long pi, ei, av;

          av = avma;
          Qx_2 = modii(sqri(A_2x_plus_B), kN);
          s = strtok(relations, " \n");
          prod_pi_ei = gun;
          while (s != NULL)
          {
            ei = atol(s);
            if (ei == 0) break;
            s = strtok(NULL, " \n");
            pi = atol(s);
            pi_ei = powmodulo(stoi(FB[pi]), stoi(ei), kN);
            prod_pi_ei = modii(mulii(prod_pi_ei, pi_ei), kN);
            s = strtok(NULL, " \n");
          }
          prod_pi_ei = modii(mulii(prod_pi_ei, Qx), kN);
          avma = av;

          if (gcmp(Qx_2, prod_pi_ei) != 0)
	    err(talker, "MPQS: wrong large prime relation found!!");
	}
#endif

	free(Qxstring);
	free(L1string);
      }
      else
      {
         /* TO BE DONE: check for double large prime */
      }
    }

    relations[0] = '\0';
    avma = av;
  }

  free(relations);
  return number_of_relations;
}

/******************************/

/**
 ** combines the large prime relations in LPREL to full
 ** relations in FREL. FREL is assumed to be open for
 ** writing / appending
 **/

typedef struct
{
  long q;
  char Y  [MPQS_STRING_LENGTH];
  char ep [MPQS_STRING_LENGTH];
} mpqs_lp_entry;

static void
mpqs_combine_exponents(long *ei, long ei_size, char *r1, char *r2)
{
  char ej [MPQS_STRING_LENGTH], ek [MPQS_STRING_LENGTH];
  char *s;
  long e, p;

  memset(ei, 0, ei_size * sizeof(long));
  strcpy(ej, r1);
  strcpy(ek, r2);

  s = strtok(ej, " \n");
  while (s != NULL)
  {
    e = atol(s);
    if (e == 0)
      break;
    s = strtok(NULL, " \n");
    p = atoi(s);
    ei[p] += e;
    s = strtok(NULL, " \n");
  }

  s = strtok(ek, " \n");
  while (s != NULL)
  {
    e = atol(s);
    if (e == 0)
      break;
    s = strtok(NULL, " \n");
    p = atoi(s);
    ei[p] += e;
    s = strtok(NULL, " \n");
  }

}

static void
mpqs_combine_large_primes(FILE *LPREL, FILE *FREL, long size_of_FB, GEN kN
#ifdef MPQS_DEBUG
  , long *FB
#endif
)
{
  char buf [MPQS_STRING_LENGTH], *s1, *s2;
  char new_relation [MPQS_STRING_LENGTH];
  mpqs_lp_entry e[50 + 1];
  long *ei, ei_size = size_of_FB + 2;
  long old_q = -1;
  long i, j, k, c = 0;

  if (fgets(buf, MPQS_STRING_LENGTH, LPREL) != NULL)
  {
    ei = (long *) gpmalloc(ei_size * sizeof(long));
    s1 = buf; s2 = strchr(s1, ' '); *s2 = '\0';
    e[0].q = atol(s1);
    s1 = s2 + 3; s2 = strchr(s1, ' '); *s2 = '\0';
    strcpy(e[0].Y, s1);
    s1 = s2 + 3;  s2 = strchr(s1, '\n'); *s2 = '\0';
    strcpy(e[0].ep, s1);

    i = 1;
    old_q = e[0].q;

    while (fgets(buf, MPQS_STRING_LENGTH, LPREL) != NULL)
    {

      s1 = buf; s2 = strchr(s1, ' '); *s2 = '\0';
      e[i].q = atol(s1);
      s1 = s2 + 3; s2 = strchr(s1, ' '); *s2 = '\0';
      strcpy(e[i].Y, s1);
      s1 = s2 + 3;  s2 = strchr(s1, '\n'); *s2 = '\0';
      strcpy(e[i].ep, s1);

      if (e[i].q == old_q)
      {
        /* we only handle at most 50 large prime relations
	   with the same q */
        if (i < 50)
          i++;
      }
      else
      {
	if (i == 1)
        {
	  old_q = e[1].q;
	  e[0].q = e[1].q;
	  strcpy(e[0].Y, e[1].Y);
	  strcpy(e[0].ep, e[1].ep);
	}
	else
	{
	  long av = avma;
	  GEN inv_q = mpinvmod(stoi(e[0].q), kN);
	  for (j = 0; j < i; j++)
          {
	    for(k = j + 1; k < i; k++)
            {
	      long av2 = avma, l;
	      GEN Y1, Y2, new_Y;
	      c++;

	      /* compute relation [j, k] */
	      mpqs_combine_exponents(ei, ei_size, e[j].ep, e[k].ep);
	      if (DEBUGLEVEL >= 6)
              {
		fprintf(stderr, "MPQS: combining\n");
		fprintf(stderr, " [%ld]  {%ld @ %s : %s}\n",
			j, e[j].q, e[j].Y, e[j].ep);
		fprintf(stderr, " [%ld]  {%ld @ %s : %s}\n",
			k, e[k].q, e[k].Y, e[k].ep);
	      }
	      Y1 = lisexpr(e[j].Y);
	      Y2 = lisexpr(e[k].Y);
	      new_Y = modii(mulii(mulii(Y1, Y2), inv_q), kN);
	      s1 = gen2str(new_Y);
	      strcpy(new_relation, s1);
	      strcat(new_relation, " :");
	      if (ei[1] % 2)
		strcat(new_relation, " 1 1");
	      for (l = 2; l < ei_size; l++)
	      {
		if (ei[l])
		{
		  sprintf(buf, " %ld %ld", ei[l], l);
		  strcat(new_relation, buf);
		}
	      }
	      strcat(new_relation, " 0");
	      if (DEBUGLEVEL >= 6)
		fprintf(stderr, "     == {%s}\n", new_relation);
	      strcat(new_relation, "\n");

#ifdef MPQS_DEBUG
	      {
		char ejk [MPQS_STRING_LENGTH];
		GEN Qx_2, prod_pi_ei, pi_ei;
		char *s;
		long pi, ei, av;

		av = avma;
		Qx_2 = modii(sqri(new_Y), kN);

		strcpy(ejk, new_relation);
		s = strchr(ejk, ':') + 2;
		s = strtok(s, " \n");

		prod_pi_ei = gun;
		while (s != NULL)
		{
		  ei = atol(s);
		  if (ei == 0) break;
		  s = strtok(NULL, " \n");
		  pi = atol(s);
		  pi_ei = powmodulo(stoi(FB[pi]), stoi(ei), kN);
		  prod_pi_ei = modii(mulii(prod_pi_ei, pi_ei), kN);
		  s = strtok(NULL, " \n");
		}
		avma = av;

		if (gcmp(Qx_2, prod_pi_ei) != 0)
		  err(talker,
		      "MPQS: combined large prime relation is false!!");
	      }
#endif

	      if (fputs(new_relation, FREL) < 0)
		bummer(FREL_str);
	      free(s1);
	      avma = av2;
	    }
	  }

	  /* reset table */
	  old_q = e[i].q;
	  e[0].q = e[i].q;
	  strcpy(e[0].Y, e[i].Y);
	  strcpy(e[0].ep, e[i].ep);
	  i = 1;
	  avma = av;
	}
      }
    }
    free(ei);
  }

  if (DEBUGLEVEL >= 4)
    fprintferr("\nMPQS: combined %ld full relations\n", c);
}

/******************************/

#ifdef LONG_IS_64BIT

#define MPQS_GAUSS_BITS 64
static unsigned long mpqs_mask_bit[]  =
{
  0x8000000000000000UL, 0x4000000000000000UL,
  0x2000000000000000UL, 0x1000000000000000UL,
  0x0800000000000000UL, 0x0400000000000000UL,
  0x0200000000000000UL, 0x0100000000000000UL,
  0x0080000000000000UL, 0x0040000000000000UL,
  0x0020000000000000UL, 0x0010000000000000UL,
  0x0008000000000000UL, 0x0004000000000000UL,
  0x0002000000000000UL, 0x0001000000000000UL,
  0x0000800000000000UL, 0x0000400000000000UL,
  0x0000200000000000UL, 0x0000100000000000UL,
  0x0000080000000000UL, 0x0000040000000000UL,
  0x0000020000000000UL, 0x0000010000000000UL,
  0x0000008000000000UL, 0x0000004000000000UL,
  0x0000002000000000UL, 0x0000001000000000UL,
  0x0000000800000000UL, 0x0000000400000000UL,
  0x0000000200000000UL, 0x0000000100000000UL,
  0x0000000080000000UL, 0x0000000040000000UL,
  0x0000000020000000UL, 0x0000000010000000UL,
  0x0000000008000000UL, 0x0000000004000000UL,
  0x0000000002000000UL, 0x0000000001000000UL,
  0x0000000000800000UL, 0x0000000000400000UL,
  0x0000000000200000UL, 0x0000000000100000UL,
  0x0000000000080000UL, 0x0000000000040000UL,
  0x0000000000020000UL, 0x0000000000010000UL,
  0x0000000000008000UL, 0x0000000000004000UL,
  0x0000000000002000UL, 0x0000000000001000UL,
  0x0000000000000800UL, 0x0000000000000400UL,
  0x0000000000000200UL, 0x0000000000000100UL,
  0x0000000000000080UL, 0x0000000000000040UL,
  0x0000000000000020UL, 0x0000000000000010UL,
  0x0000000000000008UL, 0x0000000000000004UL,
  0x0000000000000002UL, 0x0000000000000001UL
};

#else

#define MPQS_GAUSS_BITS 32
static unsigned long mpqs_mask_bit[]  =
{
  0x80000000UL, 0x40000000UL, 0x20000000UL, 0x10000000UL,
  0x08000000UL, 0x04000000UL, 0x02000000UL, 0x01000000UL,
  0x00800000UL, 0x00400000UL, 0x00200000UL, 0x00100000UL,
  0x00080000UL, 0x00040000UL, 0x00020000UL, 0x00010000UL,
  0x00008000UL, 0x00004000UL, 0x00002000UL, 0x00001000UL,
  0x00000800UL, 0x00000400UL, 0x00000200UL, 0x00000100UL,
  0x00000080UL, 0x00000040UL, 0x00000020UL, 0x00000010UL,
  0x00000008UL, 0x00000004UL, 0x00000002UL, 0x00000001UL
};

#endif

static mpqs_gauss_matrix
mpqs_gauss_create_matrix(long rows, long cols)
{
  mpqs_gauss_matrix m;
  long i, j, words = cols / MPQS_GAUSS_BITS;
  if (cols % MPQS_GAUSS_BITS)
    words++;
  m = (mpqs_gauss_row *) gpmalloc(rows * sizeof(mpqs_gauss_row));
  for (i = 0; i < rows; i++)
  {
    m[i] = (ulong *) gpmalloc(words * sizeof(ulong));
    for (j = 0; j < words; j++)
      m[i][j] = 0UL;
  }
  return m;
}

static void
mpqs_gauss_destroy_matrix(mpqs_gauss_matrix m, long rows, long cols)
{
  long i;
  (void) cols;
  for (i = 0; i < rows; i++)
    free(m[i]);
  free(m);
}

static ulong
mpqs_gauss_get_bit(mpqs_gauss_matrix m, long i, long j)
{
  return m[i][j / MPQS_GAUSS_BITS] & mpqs_mask_bit[j % MPQS_GAUSS_BITS];
}

static void
mpqs_gauss_set_bit(mpqs_gauss_matrix m, long i, long j)
{
  m[i][j / MPQS_GAUSS_BITS] |= mpqs_mask_bit[j % MPQS_GAUSS_BITS];
}

static void
mpqs_gauss_clear_bit(mpqs_gauss_matrix m, long i, long j)
{
  m[i][j / MPQS_GAUSS_BITS] &= ~mpqs_mask_bit[j % MPQS_GAUSS_BITS];
}

/* output an mpqs_gauss_matrix in PARI format */
/* TODO -- this should do a little intelligent wrapping... GN */
static void
mpqs_gauss_print_matrix(mpqs_gauss_matrix m, long rows, long cols)
{
  long i, j;
  fprintferr("\n[");
  for (i = 0; i < rows; i++)
  {
    for (j = 0; j < cols - 1; j++)
    {
      if (mpqs_gauss_get_bit(m, i, j))
        fprintferr("1, ");
      else
        fprintferr("0, ");
    }

    if (mpqs_gauss_get_bit(m, i, j))
      fprintferr("1");
    else
      fprintferr("0");
    if (i != rows - 1)
      fprintferr("; ");
  }
  fprintferr("]\n");
}

/* x ^= y : row addition over F_2 */
static void
mpqs_gauss_add_rows(mpqs_gauss_row y, mpqs_gauss_row x, long k, long n)
{
  long i, q, r;
  n = n - k;
  r = n % 8; q = n - r + k; i = 0 + k;
  for (; i < q; i += 8)
  {
    x[0+i] ^= y[0+i]; x[1+i] ^= y[1+i];
    x[2+i] ^= y[2+i]; x[3+i] ^= y[3+i];
    x[4+i] ^= y[4+i]; x[5+i] ^= y[5+i];
    x[6+i] ^= y[6+i]; x[7+i] ^= y[7+i];
  }
  switch (r)
  {
    case 7: x[i] ^= y[i], i++;
    case 6: x[i] ^= y[i], i++;
    case 5: x[i] ^= y[i], i++;
    case 4: x[i] ^= y[i], i++;
    case 3: x[i] ^= y[i], i++;
    case 2: x[i] ^= y[i], i++;
    case 1: x[i] ^= y[i], i++;
    default: ;
  }
}

/* read and create an mpqs_gauss_matrix from a relation file FREL
   rows = size_of_FB, cols = rel */
static mpqs_gauss_matrix
mpqs_gauss_read_matrix(long rows, long cols)
{
  mpqs_gauss_matrix m;
  long i = 0, e, p;
  char buf[MPQS_STRING_LENGTH], *s;

  FILE *fp;
  m = mpqs_gauss_create_matrix(rows, cols);
  fp = mpqs_open_file(FREL_str, "r");
  while (fgets(buf, MPQS_STRING_LENGTH, fp) != NULL)
  {
    s = strchr(buf, ':') + 2;
    s = strtok(s, " \n");
    while (s != NULL)
    {
      e = atol(s);
      if (e == 0)
	break;
      s = strtok(NULL, " \n");
      p = atol(s);
      if (e & 1)
	mpqs_gauss_set_bit(m, p - 1, i);
      s = strtok(NULL, " \n");
    }
    i++;
  }
  mpqs_close_file(fp);
  return m;
}

/* compute the kernel of an mpqs_gauss_matrix over F_2
   uses the algorithm described by Knuth, Vol. 2 as modified by Henri Cohen
   and optimized by Christian Batut, Thomas Papanikolaou and Xavier Roblot.
   m = rows, n = cols */
static mpqs_gauss_matrix
mpqs_kernel(mpqs_gauss_matrix x, long m, long n, long *rank)
{
  long k, i, j, t, r;
  mpqs_gauss_matrix ker_x;
  mpqs_gauss_indices c, d;
  long words = n / MPQS_GAUSS_BITS;
  if (n % MPQS_GAUSS_BITS)
    words++;

  c = (long *) gpmalloc(m * sizeof(long));
  for (k = 0; k < m; k++) c[k] = -1;

  d = (long *) gpmalloc(n * sizeof(long));
  r = 0;

  for (k = 0; k < n; k++)
  {
    j = 0;
    while (j < m && (c[j] >= 0 || mpqs_gauss_get_bit(x, j, k) == 0))
      j++;

    if (j == m)
    {
      d[k] = -1; /* no pivot found in column k */
      r++;       /* column k is a vector of the kernel */
    }
    else
    {
      d[k] = j; /* the pivot in column k is at row j */
      c[j] = k; /* a pivot at position j was found in column k */
      for (t = 0; t < m; t++)
      {
	if (t != j)
	{
	  if (mpqs_gauss_get_bit(x, t, k))
	    mpqs_gauss_add_rows(x[j], x[t], k / MPQS_GAUSS_BITS, words);
        }
      }
    }
  }

  ker_x = mpqs_gauss_create_matrix(n, r);

  for (j = k = 0; j < r; j++, k++)
  {
    while (d[k] != -1) k++;
    for (i = 0; i < k; i++)
    {
      if (d[i] != -1)
      {
	if (mpqs_gauss_get_bit(x, d[i], k))
	  mpqs_gauss_set_bit(ker_x, i, j);
	else
	  mpqs_gauss_clear_bit(ker_x, i, j);
      }
      else
	mpqs_gauss_clear_bit(ker_x, i, j);
    }
    mpqs_gauss_set_bit(ker_x, k, j);
  }

  free(c);
  free(d);

  *rank = r;
  return ker_x;
}

/******************************/

static GEN
mpqs_add_relation(GEN Y_prod, GEN N_or_kN, long *ei, char *r1)
{
  GEN Y, res;
  char relation [MPQS_STRING_LENGTH];
  char *s;
  long av = avma, e, p;

  strcpy(relation, r1);
  s = strchr(relation, ':') - 1;
  *s = '\0';
  Y = lisexpr(relation);
  res = modii(mulii(Y_prod, Y), N_or_kN);

  s = s + 3;
  s = strtok(s, " \n");
  while (s != NULL)
  {
    e = atol(s);
    if (e == 0)
      break;
    s = strtok(NULL, " \n");
    p = atoi(s);
    ei[p] += e;
    s = strtok(NULL, " \n");
  }
  return gerepileupto(av,res);
}

static char*
mpqs_get_relation(long pos, FILE *fp)
{
  static char buf [MPQS_STRING_LENGTH];
  if (fseek(fp, pos, SEEK_SET))
  {
    mpqs_close_file(fp);
    /* may also fail, but we're giving up anyway... GN */
    err(talker, "can\'t seek relations file");
  }
  if (fgets(buf, MPQS_STRING_LENGTH, fp) == NULL)
  {
    mpqs_close_file(fp);
    err(talker, "relations file truncated?!");
  }
  return buf;
}

int miller(GEN n, long k);	/* in src/basemath/ifactor1.c */
#define isprobableprime(n) (miller((n),17))

static GEN
mpqs_solve_linear_system(GEN kN, GEN N, long rel, long *FB, long size_of_FB)
{
  FILE *fp;
  GEN X, Y_prod, N_or_kN, D1, base, res, new_res;
  char buf [MPQS_STRING_LENGTH];
  long av=avma, av2, av3, lim, tetpil;
  long *fpos, *ei;
  long i, j, H_cols, H_rows;
  long res_last, res_next, res_size, res_max;
  mpqs_gauss_matrix m, ker_m;
  long done, flag, mask, rank;

#ifdef MPQS_DEBUG
  N_or_kN = kN;
#else
  N_or_kN = (DEBUGLEVEL >= 4 ? kN : N);
#endif /* --GN */
  fp = mpqs_open_file(FREL_str, "r");
  fpos = (long *) gpmalloc(rel * sizeof(long));
  for (i = 0; i < rel; i++)
  {
    if ((fpos[i] = ftell(fp)) < 0)
    {
      mpqs_close_file(fp);
      err(talker, "ftell error on full relations file");
    }
    if (fgets(buf, MPQS_STRING_LENGTH, fp) == NULL)
    {
      mpqs_close_file(fp);
      err(talker, "full relations file truncated?!");
    }
  }
  mpqs_close_file(fp);
  /* somewhat inefficient... we could rewind it, and m'gauss_read_matrix()
   * could start and end with the file already open  (and then we'll rewind
   * it again, or some such thing) --GN */

  m = mpqs_gauss_read_matrix(size_of_FB+1, rel);
  if (DEBUGLEVEL >= 7)
  {
    fprintferr("\\\\ MATRIX READ BY MPQS\nFREL=");
    mpqs_gauss_print_matrix(m, size_of_FB+1, rel);
    fprintferr("\n");
  }

  ker_m = mpqs_kernel(m, size_of_FB+1, rel, &rank);
  if (DEBUGLEVEL >= 4)
  {
    if (DEBUGLEVEL >= 7)
    {
      fprintferr("\\\\ KERNEL COMPUTED BY MPQS\nKERNEL=");
      mpqs_gauss_print_matrix(ker_m, rel, rank);
      fprintferr("\n");
    }
    /* put this here where the poor user has a chance of seeing it: */
    fprintferr("MPQS: Gauss done: kernel has rank %ld, taking gcds...\n",
	       rank);
  }

  H_rows = rel;
  H_cols = rank;

  /* If the computed kernel turns out to be trivial, fail gracefully;
     main loop may try to find some more relations --GN */
  if (H_cols == 0)
  {
    if (DEBUGLEVEL >= 3)
      err(warner, "MPQS: no solutions found from linear system solver");
    /* the FREL file happens to be closed at this point, and ei has not
       yet been allocated */
    mpqs_gauss_destroy_matrix(m, size_of_FB+1, rel);
    mpqs_gauss_destroy_matrix(ker_m, rel, rank);
    free(fpos);
    avma = av;
    return NULL; /* no factors found */
  }

  /* If the rank is r, we can expect up to 2^r pairwise coprime factors,
     but it may well happen that a kernel basis vector contributes nothing
     new to the decomposition.  We will allocate room for max(2,r) factors
     initially  (certainly adequare when one or two basis vectors work),
     adjusting this down at the end to what we actually found, or up if
     we are very lucky and find more factors.  In the upper half of our
     vector, we store information about which factors we know to be composite
     (zero) or believe to be composite ((long)NULL which is just 0) or suspect
     to be prime (un), or an exponent (deux or some t_INT) if it is a proper
     power */
  av2 = avma; lim = (av2+bot)>>1;
  if (rank > BITS_IN_LONG - 2)
    res_max = VERYBIGINT;	/* this, unfortunately, is the common case */
  else
    res_max = 1<<rank;		/* max number of factors we can hope for */
  res_size = rank;		/* no. of factors we can store in this res */
  if (res_size < 2) res_size = 2;
  res = cgetg(2*res_size+1, t_VEC);
  for (i=2*res_size; i; i--) res[i] = 0; /* make it safe for gcopy() */
  res_next = res_last = 1;

  /* re-opens FREL file a third time for reading (see above) --GN */
  fp = mpqs_open_file(FREL_str, "r");
  /* Recomputation of N from kN removed --GN */

  /* use PARI idiom -- create results on the stack;  only clean up when we
     absolutely have to  (not very likely;  numbers treated by mpqs() are
     short after all, compared to a typical PARI stack) --GN */
#if 0
  Y_prod = cgeti(lg(N_or_kN) + 1);
  X = cgeti(lg(N_or_kN) + 1);
#endif
  ei = (long *) gpmalloc((size_of_FB + 2) * sizeof(long));

  for (i = 0; i < H_cols; i++)	/* loop over basis of kernel */
  {
#if 0
    affii(gun, Y_prod);
#else
    X = Y_prod = gun;
#endif
    memset(ei, 0, (size_of_FB + 2) * sizeof(long));

    for (j = 0; j < H_rows; j++)
    {
      if (mpqs_gauss_get_bit(ker_m, j, i))
	Y_prod = mpqs_add_relation(Y_prod, N_or_kN, ei,
				   mpqs_get_relation(fpos[j], fp));
    }
#if 0
    affii(gun, X);
#endif

    for (j = 2; j <= (size_of_FB + 1); j++)
    {
      if (ei[j] & 1)
      {
	mpqs_close_file(fp);
	mpqs_gauss_destroy_matrix(m, size_of_FB+1, rel);
	mpqs_gauss_destroy_matrix(ker_m, rel, rank);
	free(ei); free(fpos);
	fprintferr("MPQS: the combination of the relations is a nonsquare\n");
	err(bugparier, "factoring (MPQS)");
      }
      if (ei[j])
      {
	X = modii(mulii(X, powmodulo(stoi(FB[j]), stoi(ei[j]>>1), N_or_kN)),
		  N_or_kN);
      }
    }
    if (
#ifdef MPQS_DEBUG
	1 ||
#endif
	DEBUGLEVEL >= 1)
    {
      if (signe(modii(subii(sqri(X), sqri(Y_prod)), N_or_kN)))
      {
	mpqs_close_file(fp);
	mpqs_gauss_destroy_matrix(m, size_of_FB+1, rel);
	mpqs_gauss_destroy_matrix(ker_m, rel, rank);
	free(ei); free(fpos);
	fprintferr("MPQS: X^2 - Y^2 != 0 mod %s\n",
		   (N_or_kN == kN ? "kN" : "N"));
	fprintferr("\tindex i = %ld\n", i);
	err(bugparier, "factoring (MPQS)");
      }
    }
    /* if res_next < 3, we still haven't decomposed the original N, and
       will want both a gcd and its cofactor, overwriting N.  Note that
       gcd(X+Y_prod,N)==1 forces gcd(X-Y_prod,N)==N, so we can skip X-Y_prod
       in such cases */
    done = 0;			/* (re-)counts probably-prime factors
				   (or powers whose bases we don't want to
				   handle any further) */
    if (res_next < 3)
    {
      D1 = mppgcd(addii(X,Y_prod),N);
      if (!is_pm1(D1))
      {
	if ((flag = egalii(D1, N))) /* assignment */
				/* this one's useless, try the other one */
	  D1 = mppgcd(subii(X,Y_prod),N);
	if (!flag || (!is_pm1(D1) && !egalii(D1,N)))
	{			/* got something that works */
          if (DEBUGLEVEL >= 5)
            fprintferr("MPQS: splitting N after %ld kernel vector%s\n",
                       i+1, (i? "s" : ""));
	  (void)mpdivis(N, D1, N); /* divide N in place */
	  res[1] = (long)N;	/* we'll gcopy this anyway before exiting */
	  res[2] = (long)D1;
	  res_last = res_next = 3;
	  if (res_max == 2) break; /* two out of two possible factors seen */
	  mask = 7;
	  /* (actually, 5th/7th powers aren't really worth the trouble.  OTOH
	     once we have the hooks for dealing with cubes, higher powers can
	     be handled essentially for free) */
	  if (isprobableprime(N)) { done++; res[res_size+1] = un; }
	  else if (carrecomplet(N, &base))
	  {			/* squares could cost us a lot of time */
	    affii(base, N); done++; res[res_size+1] = deux;
	    if (DEBUGLEVEL >= 5)
	      fprintferr("MPQS: decomposed a square\n");
	  }
	  else if ((flag = is_odd_power(N, &base, &mask))) /* assignment */
	  {
	    affii(base, N); done++; res[res_size+1] = (long)(stoi(flag));
	    if (DEBUGLEVEL >= 5)
	      fprintferr("MPQS: decomposed a %s\n",
			 (flag == 3 ? "cube" :
			  (flag == 5 ? "5th power" : "7th power")));
	  }
	  else res[res_size+1] = zero; /* known composite */
	  mask = 7;
	  if (isprobableprime(D1)) { done++; res[res_size+2] = un; }
	  else if (carrecomplet(D1, &base))
	  {
	    res[2] = (long)base;
	    done++; res[res_size+2] = deux;
	    if (DEBUGLEVEL >= 5)
	      fprintferr("MPQS: decomposed a square\n");
	  }
	  else if ((flag = is_odd_power(D1, &base, &mask))) /* assignment */
	  {
	    res[2] = (long)base;
	    done++; res[res_size+2] = (long)(stoi(flag));
	    if (DEBUGLEVEL >= 5)
	      fprintferr("MPQS: decomposed a %s\n",
			 (flag == 3 ? "cube" :
			  (flag == 5 ? "5th power" : "7th power")));
	  }
	  else res[res_size+2] = zero; /* known composite */
	  if (done == 2) break;	/* both factors look prime or were powers */
	  if (DEBUGLEVEL >= 5)
	    fprintferr("MPQS: got two factors, looking for more...\n");
	  /* and loop (after collecting garbage if necessary) */
	} /* else loop to next kernel basis vector */
      } /* D1!=1, nothing to be done when ==1 */
    }
    else			/* we already have factors */
    {
      GEN X_plus_Y = addii(X, Y_prod);
      GEN X_minus_Y = NULL;	/* will only be computed when needed */
      for (j=1; j < res_next; j++)
      {				/* loop over known-composite factors */
	if (res[res_size+j] && res[res_size+j] != zero)
	{
	  done++; continue;	/* skip probable primes etc */
	}
	/* (actually, also skip square roots of squares etc.  They are
	   necessarily a lot smaller than the original N, and should
	   be easy to deal with later) */
	av3 = avma;
	D1 = mppgcd(X_plus_Y, (GEN)(res[j]));
	if (is_pm1(D1)) continue; /* this one doesn't help us */
	if ((flag = egalii(D1, (GEN)(res[j]))))
	{			/* bad one, try the other */
	  if (!X_minus_Y) X_minus_Y = subii(X, Y_prod);
	  D1 = mppgcd(X_minus_Y, (GEN)(res[j]));
	}
	if (!flag || (!is_pm1(D1) && !egalii(D1, (GEN)(res[j]))))
	{			/* got one which splits this factor */
          if (DEBUGLEVEL >= 5)
            fprintferr("MPQS: resplitting a factor after %ld kernel vectors\n",
                       i+1);      /* always plural */
	  /* first make sure there's room for another factor */
	  if (res_next > res_size)
	  {			/* need to reallocate (_very_ rare case) */
	    long i1, new_size = 2*res_size;
	    GEN new_res;
	    if (new_size > res_max) new_size = res_max;
	    new_res = cgetg(2*new_size+1, t_VEC);
	    for (i1=2*new_size; i1>=res_next; i1--) new_res[i1] = 0;
	    for (i1=1; i1<res_next; i1++)
	    {
	      new_res[i1] = res[i1]; /* factors */
	      new_res[new_size+i1] = res[res_size+i1]; /* primality tags */
	    }
	    res = new_res; res_size = new_size;	/* res_next unchanged */
	  }
	  /* now there is room; divide into existing factor and store the
	     new gcd */
	  (void)mpdivis((GEN)(res[j]), D1, (GEN)(res[j]));
	  res[res_next] = (long)D1;
	  if (++res_next > res_max) break; /* all possible factors seen */
	  mask = 7;		/* check for 3rd and 5th powers */
	  if (isprobableprime((GEN)(res[j])))
	  {
	    done++; res[res_size+j] = un;
	  }
	  else if (carrecomplet((GEN)(res[j]), &base))
	  {
	    /* we no longer bother preserving the cloned N or what is left
	       of it when we hit it here */
	    res[j] = (long)base;
	    done++; res[res_size+j] = deux;
	    if (DEBUGLEVEL >= 5)
	      fprintferr("MPQS: decomposed a square\n");
	  }
	  else if ((flag = is_odd_power((GEN)(res[j]), &base, &mask))) /* = */
	  {
	    res[j] = (long)base;
	    done++; res[res_size+j] = (long)stoi(flag);
	    if (DEBUGLEVEL >= 5)
	      fprintferr("MPQS: decomposed a %s\n",
			 (flag == 3 ? "cube" :
			  (flag == 5 ? "5th power" : "7th power")));
	  }
	  else res[res_size+j] = zero; /* known composite */
	  mask = 7;
	  if (isprobableprime(D1))
	  {
	    done++; res[res_size+res_last] = un;
	  }
	  else if (carrecomplet(D1, &base))
	  {
	    res[res_last] = (long)base;
	    done++; res[res_size+res_last] = deux;
	    if (DEBUGLEVEL >= 5)
	      fprintferr("MPQS: decomposed a square\n");
	  }
	  else if ((flag = is_odd_power(D1, &base, &mask))) /* assignment */
	  {
	    res[res_last] = (long)base;
	    done++; res[res_size+res_last] = (long)stoi(flag);
	    if (DEBUGLEVEL >= 5)
	      fprintferr("MPQS: decomposed a %s\n",
			 (flag == 3 ? "cube" :
			  (flag == 5 ? "5th power" : "7th power")));
	  }
	  else res[res_size+res_last] = zero; /* known composite */
	} /* else it didn't help on this factor, try the next one... */
	else avma = av3;	/* ...after destroying the gcds */
      }	/* loop over known composite factors */
      if (res_next > res_last)
      {
	if (DEBUGLEVEL >= 5 && done < res_last)
	  fprintferr("MPQS: got %ld factors%s\n",
		     res_last,
		     (done < res_last ? ", looking for more..." : ""));
	res_last = res_next;
      }
      /* we should break out of the outer loop when we have seen res_max
	 factors, and also when all current factors are probable primes */
      if (res_next > res_max || done == res_next - 1) break;
      /* else continue loop over kernel basis */
    } /* end case of further splitting of existing factors */
    /* garbage collection */
    if (low_stack(lim, (av2+bot)>>1))
    {
      long i1;
      if(DEBUGMEM>1) err(warnmem,"mpqs_solve_linear_system");
      tetpil=avma;
      /* gcopy would have a problem with our NULL pointers... */
      new_res = cgetg(lg(res), t_VEC);
      for (i1=2*res_size; i1>=res_next; i1--) new_res[i1] = 0;
      for (i1=1; i1<res_next; i1++)
      {
	new_res[i1] =
	  (isonstack((GEN)(res[i1])) ? licopy((GEN)(res[i1])) : res[i1]);
	new_res[res_size+i1] = res[res_size+i1];
      }
      res = gerepile(av2, tetpil, new_res);
      /* discard X and Y_prod */
    }
  } /* for (loop over kernel basis) */

  if (res_next < 3)		/* still no factors seen */
  {
    mpqs_close_file(fp);
    mpqs_gauss_destroy_matrix(m, size_of_FB+1, rel);
    mpqs_gauss_destroy_matrix(ker_m, rel, rank);
    free(ei); free(fpos);
    avma = av;
    return NULL; /* no factors found */
  }
  /* normal case:  convert internal format to ifac format as described in
     src/basemath/ifactor1.c  (triples of components: value, exponent, class
     [unknown, known composite, known prime]),  clean up and return result */
  mpqs_close_file(fp);
  mpqs_gauss_destroy_matrix(m, size_of_FB+1, rel);
  mpqs_gauss_destroy_matrix(ker_m, rel, rank);
  free(ei); free(fpos);
  res_last = res_next - 1;	/* number of distinct factors found */
  new_res = cgetg(3*res_last + 1, t_VEC);
  if (DEBUGLEVEL >= 6)
    fprintferr("MPQS: wrapping up vector of %ld factors\n", res_last);
  for (i=1,j=1; i <= res_last; i++)
  {
    new_res[j++] =		/* value of factor */
      isonstack((GEN)(res[i])) ? licopy((GEN)(res[i])) : res[i];
    flag = res[res_size+i];
    new_res[j++] =		/* exponent */
      flag ?			/* flag was zero or un or ... */
	(flag == zero ? un :
	 (isonstack((GEN)flag) ? licopy((GEN)flag) : flag)
	 ) :
	   un;			/* flag was (long)NULL */
    new_res[j++] =		/* class */
      flag == zero ? zero :	/* known composite */		
	(long)NULL;		/* base of power or suspected prime --
				   mark as `unknown' */
    if (DEBUGLEVEL >= 6)
      fprintferr("\tpackaging %ld: %Z ^%ld (%s)\n",
		 i,
		 /* (long) anyway */ res[i],
		 itos((GEN)(new_res[j-2])),
		 (flag == zero ? "comp." : "unknown"));
  }
  return gerepileupto(av, new_res);
}

/******************************/

#define max_percentage 200	/* give up when nothing found after twice
				   the required number of relations has
				   been computed (n might be a prime power
				   or the parameters might be exceptionally
				   unfortunate for it) */

/**
 ** Factors N using the self-initializing multipolynomial quadratic sieve
 ** (SIMPQS).  Returns one of the two factors.  The factor as well as the
 ** cofactor may still be composite.
 **/

GEN
mpqs(GEN N)
{
  ulong decimal_digits;		/* number of decimals of N */
  long tc;			/* number of candidates found in one
				   iteration */
  long t;			/* number of full relations found in one
				   iteration */
  long tp;			/* number of full relations found in one
				   iteration after combining large prime
				   relations */
  long k;			/* the multiplier */
  GEN kN;			/* k * N */
  double sqrt_kN;		/* double approximation of sqrt(k*N) */
  long *FB;			/* the factor base */
  ulong size_of_FB;		/* the size of the factor base */
  double log_multiplier;	/* maps logarithms of FB elements to
				   unsigned chars */
  double tolerance;		/* used in sieving with logarithms */
  unsigned char *log_FB;	/* logarithms of FB elements mapped to
				   unsigned chars */
  unsigned char *sieve_array;	/* the sieve array */
  unsigned char *sieve_array_end; /* the end of the sieve array */
  ulong M;			/* the sieve interval size [-M, M] */
  ulong starting_sieving_index;	/* the first element in the FB used for
				   sieving */
  long last_prime_in_FB;	/* the biggest prime in FB */
  long *sqrt_mod_p_kN;		/* array containing sqrt(K*N) mod p_i,
				   where p_i are FB elements */
  GEN A, B, inv_A4;		/* coefficients of polynomials in (SI) */
  long *Q_prime_glob, *Q_prime;	/* used for self initialization (SI) */
  GEN *H_i;			/* will be used to compute the consecutive
				   B during (SI) */
  long start_index_FB_for_A;	/* the first index in the FB used for
				   computing A during (SI) */
  ulong no_of_primes_in_A;	/* number of primes dividing A
				   during (SI) */
  ulong total_no_of_primes_for_A; /* number of primes used for finding A
				     during (SI) */
  long max_no_of_B;		/* how many B's to use before choosing
				   a new A during (SI) */
  long index_i;			/* denotes the current pair of coeffs
				   (A, B_i). If i == 0, a new A is
				   generated */
  ulong bin_index;		/* used for choosing the prime factors for
				   for the A coeffs during (SI)'s
				   mpqs_compute_coeff */
  long i, p;			/* counters */

  /* Let (z0, z1) be the roots of Q(x) = A x^2 + Bx +C mod p_i;
     then we know that
     Q(z1 + i k) == 0 mod p_i and Q(z2 + i k) == 0 mod p_i */

  long *start_1;		/* stores the position where p_i
				   divides Q(x) for the first time,
				   using z1 */
  long *start_2;		/* same, but using z2 */
  long *inv_A2;			/* inv_A2[i] = 1/(2*A) mod p_i */
  long **inv_A_H_i;		/* inv_A_H_i[i][j] = 1/A H_i[i] mod p_j */
  long *candidates;		/* positions in the sieve_array which
				   probably factor over the FB */
  long sort_interval = 10;
  long total_full_relations = 0;
  long total_partial_relations = 0;
  FILE *FREL;
  FILE *LPREL;
  FILE *LPNEW;
  long percentage = 0;
  long total_candidates_number = 0;
  long useless_candidates = 0;
  long vain_iterations = 0;
  long good_iterations = 0;
  long iterations = 0;
  long av = avma;

  if (DEBUGLEVEL >= 4)
  {
    (void) timer2();		/* clear timer */
    fprintferr("MPQS: number to factor N = %Z\n", (long)N);
  }

  all_clean = 0;		/* activate clean-up trap */

  /* pick up PARI's current differences-of-small-primes array */
  /* unless we already have our own.  We must then reset it to NULL */
  /* so we will not use a stale one the next time we are called */
  if (!mpqs_diffptr) mpqs_diffptr = diffptr;

  decimal_digits = strlen(gen2str(N));

  if (decimal_digits >= 100)
  {
    err(warner,
	"MPQS: number too big to be factored with MPQS, giving up");
    if (mpqs_diffptr == diffptr) mpqs_diffptr = NULL;
    avma = av;
    all_clean = 1;
    return NULL;
  }

  if (DEBUGLEVEL >= 4)
    fprintferr("MPQS: factoring number of %ld decimal digits\n",
	       decimal_digits);

  if (decimal_digits >= 70)
    err(warner,
	"MPQS: the factorization of this number will take several hours");

  k = mpqs_find_k(N, 5);
  if (DEBUGLEVEL >= 5)
    fprintferr("MPQS: found multiplier %ld for N\n", k);

  kN = mulis(N, k);
  decimal_digits = strlen(gen2str(kN));

  if (DEBUGLEVEL >= 5)
  {
    fprintferr("MPQS: kN = %Z\n", (long)kN);
    fprintferr("MPQS: kN has %ld decimal digits\n",
	       decimal_digits);
  }

  mpqs_read_parameters(decimal_digits, &tolerance, &M, &size_of_FB,
                       &no_of_primes_in_A, &total_no_of_primes_for_A,
                       &max_no_of_B, &starting_sieving_index);

  if (DEBUGLEVEL >= 4)
  {
    fprintferr("MPQS: sieving interval = [%ld, %ld]\n",
	       -(long)M, M);
    fprintferr("MPQS: size of factor base = %ld\n",
	       size_of_FB);
    fprintferr("MPQS: initial sieving index = %ld\n",
	       starting_sieving_index);
  }
  sieve_array =
    (unsigned char *) gpmalloc((M << 1) * sizeof(unsigned char));
  sieve_array_end = sieve_array + (M << 1) - 1;

  if (DEBUGLEVEL >= 5)
    fprintferr("MPQS: creating factor base FB of size = %ld\n",
	       size_of_FB);

  FB = mpqs_create_FB(size_of_FB, kN, k, &p);

  if (p != 0)
  {
    free(FB);
    free(sieve_array);
    if (DEBUGLEVEL >= 4)
      fprintferr("\nMPQS: found factor = %ld whilst creating factor base\n",
		 p);
    if (mpqs_diffptr == diffptr) mpqs_diffptr = NULL;
    avma = av;
    /* TODO: free the FB etc!!! */
    all_clean = 1;
    return stoi(p);
  }

  if (DEBUGLEVEL >= 5)
  {
    fprintferr("MPQS: computing logarithm approximations for p_i in FB\n");
    fprintferr("MPQS: computing sqrt(k*N) mod p_i\n");
  }
  last_prime_in_FB = FB[FB[0] + 1];
  log_multiplier = (2 * (unsigned char) (0.5 * log2 (gtodouble(kN)) +
                           log2 ((double) M) - tolerance *
                           log2 ((double) last_prime_in_FB)));
  log_multiplier = 127.0 / log_multiplier;

  log_FB =
    (unsigned char *) gpmalloc((size_of_FB + 2) * sizeof(unsigned char));
  sqrt_mod_p_kN =
    (long *) gpmalloc((size_of_FB + 2) * sizeof(long));

  for (i = 2; i < size_of_FB + 2; i++)
  {
    long av1 = avma;
    p = FB[i];

    /* compute the approximations of the logarithms of p_i */
    log_FB[i] =
      (unsigned char) (log_multiplier * log2 ((double) p) * 2);

    /* compute the x_i such that x_i^2 = (kN % p_i) mod p_i */
    sqrt_mod_p_kN[i] = itos(mpsqrtmod(modis(kN, p), stoi(p)));
    avma = av1;
  }

  if (DEBUGLEVEL >= 5)
    fprintferr("MPQS: allocating arrays for self-initialization\n");

  /* the size of coefficient A should be approximately */
  /* sqrt(kN)/M, so the size of the primes p dividing */
  /* A should be approximately (sqrt(kN/M))^(1/no_of_primes_in_A) */

  /* compute expected size of coefficients, which also defines the tolerance
     for the logarithmic sieve */
  sqrt_kN = sqrt(gtodouble(kN));
  tolerance = sqrt_kN / M;
  tolerance =  pow(2.0, log2(tolerance) / no_of_primes_in_A);
  if (tolerance > FB[size_of_FB - 1]) /* ??? */
    err(talker, "MPQS: number of prime factors in A is too small");

  /* choose primes in FB to use for building the A coefficient */
  start_index_FB_for_A = 2;
  while (FB[start_index_FB_for_A] < tolerance)
    start_index_FB_for_A++;

  if (start_index_FB_for_A > 7)
    start_index_FB_for_A -= (no_of_primes_in_A >> 1);

  /* Ensure that mthe ultiplier k does not occur in the factor base */
  if (k >= FB[start_index_FB_for_A])
  {
    do
      start_index_FB_for_A++;
    while (k >= FB[start_index_FB_for_A]);
  }

  /* total_no_of_primes_for_A consecutive primes
   * p[start_index_FB_for_A], ... are chosen from the factor base FB */

  /* collect prime numbers from which we will build the A coefficents */
  Q_prime =
    (long *) gpmalloc(no_of_primes_in_A * sizeof(long));
  Q_prime_glob =
    (long *) gpmalloc(total_no_of_primes_for_A * sizeof(long));
  for (i = 0; i < total_no_of_primes_for_A; i++)
    Q_prime_glob[i] = FB[start_index_FB_for_A + i + 1];


  /* used for choosing the correct A coeffs in (SI)'s
     mpqs_compute_coeff */
  bin_index = 0;

  /* will be used to compute the consecutive B during
     self-initialization */
  H_i = (GEN *) gpmalloc(no_of_primes_in_A * sizeof(GEN));
  for (i = 0; i < no_of_primes_in_A; i++)
    H_i[i] = cgeti(2 * total_no_of_primes_for_A);

  inv_A_H_i =
    (long **) gpmalloc(total_no_of_primes_for_A * sizeof(long *));
  for (i = 0; i < total_no_of_primes_for_A; i++)
    inv_A_H_i[i] = (long *) gpmalloc((size_of_FB + 2) * sizeof(long));

  start_1 = (long *) gpmalloc((size_of_FB + 2) * sizeof(long));
  start_2 = (long *) gpmalloc((size_of_FB + 2) * sizeof(long));

  /* the next will hold 1/(2*A) mod p_i, p_i in FB */
  inv_A2 = (long *) gpmalloc((size_of_FB + 2) * sizeof(long));

  candidates = (long *) gpmalloc(MPQS_CANDIDATE_ARRAY_SIZE * sizeof(long));

/* CLEAN UP define */

#define CLEANUP() \
  free(FB); \
  free(sieve_array); \
  free(log_FB); \
  free(sqrt_mod_p_kN); \
  free(Q_prime); \
  free(Q_prime_glob); \
  free(H_i); \
  for (i = 0; i < total_no_of_primes_for_A; i++) \
    free(inv_A_H_i[i]); \
  free(inv_A_H_i); \
  free(start_1); \
  free(start_2); \
  free(inv_A2); \
  free(candidates);

  /*
   * main loop which
   * - computes polynomials and their zeros (self initialization)
   * - does the sieving
   * - tests candidates of the sieve array
   */

  /* denotes the current pair of coeffs (A, B_i). If i == 0 a new A is
    generated */
  index_i = -1;

  /* the coefficients of the polynomial currently used for sieving */
  /* A will be at most as long as the product of
     no_of_primes_in_A word-sized primes (was total_..., GN) */
  /* B needs at most total_no_of_primes_for_A + 1 words [CHECK] --GN */
  /* Take the double is just for safety ... */
  A = cgeti(2 * no_of_primes_in_A);
  B = cgeti(2 * total_no_of_primes_for_A + 1);

  /* 1/4A mod kN */
  inv_A4 = cgeti(lg(kN));

  if (DEBUGLEVEL >= 5)
    fprintferr("MPQS: starting main loop\n");
  /* compute names for the temp files we'll need */
  FREL_str  = mpqs_get_filename("FREL");
  LPREL_str = mpqs_get_filename("LPREL");
  LPNEW_str = mpqs_get_filename("LPNEW");
  LPTMP_str = mpqs_get_filename("LPTMP");

  /* FREL is to be opened in Append mode, but as such this leads to problems
   * when we are getting interrupted.  So we truncate it first. --TP,GN */
  FREL = mpqs_open_file(FREL_str, "w");
  mpqs_close_file(FREL);
  FREL = mpqs_open_file(FREL_str, "a");
  LPREL = mpqs_open_file(LPREL_str, "w");
  /* Just for truncating, too */
  mpqs_close_file(LPREL);
  LPNEW = mpqs_open_file(LPNEW_str, "w");

  for(;;)
  {
    iterations++;
    /* when all of the B's have already been used, choose new A ;
       this is indicated by setting index_i to 0 */
    if (index_i == (max_no_of_B - 1))
      index_i = 0;
    else
      index_i++;

    /* self initialization: compute polynomial and its zeros */

    mpqs_self_init(A, B, kN, FB, sqrt_mod_p_kN, start_1, start_2,
		   no_of_primes_in_A, total_no_of_primes_for_A,
		   M, inv_A_H_i, Q_prime, Q_prime_glob, H_i,
		   index_i, start_index_FB_for_A, inv_A2, inv_A4,
		   &bin_index);
    if (bin_index >= (1ul << total_no_of_primes_for_A)) /* wraparound */
    {
      /* TODO: increase some parameters.  For the moment, simply give up */
      CLEANUP();
      mpqs_close_file(FREL);
      mpqs_close_file(LPNEW);	/* LPREL is closed at this point */
      if (unlink(FREL_str))
	err(warner, "can\'t remove file %s", FREL_str);
      if (unlink(LPREL_str))
	err(warner, "can\'t remove file %s", LPREL_str);
      if (unlink(LPNEW_str))
	err(warner, "can\'t remove file %s", LPNEW_str);
      if (mpqs_diffptr == diffptr) mpqs_diffptr = NULL;
      all_clean = 1;
      avma=av; return NULL;
    }
    if (DEBUGLEVEL >= 6)
    {
      if (!index_i)
	fprintferr("MPQS: chose prime pattern 0x%lX for A\n",
		   bin_index);
      if (signe(B) < 0)
      {
	setsigne(B,1);
	fprintferr("MPQS: chose Q_%ld(x) = %Z x^2 - %Z x + C\n",
		   index_i, (long)A, (long)B);
	setsigne(B,-1);
      }
      else
      {
	fprintferr("MPQS: chose Q_%ld(x) = %Z x^2 + %Z x + C\n",
		   index_i, (long)A, (long)B);
      }
    }

    mpqs_sieve(FB, log_FB, start_1, start_2, sieve_array,
	       sieve_array_end, M, starting_sieving_index);

    tc = mpqs_eval_sieve(sieve_array, M, candidates);
    total_candidates_number += tc;
    if (DEBUGLEVEL >= 6)
      fprintferr("MPQS: found %lu candidate%s\n",
		 tc, (tc==1? "" : "s"));

    if (tc)
    {
      good_iterations++;
      t = mpqs_eval_candidates(A, inv_A4, B, kN, k, sqrt_kN, FB,
			       start_1, start_2,
			       no_of_primes_in_A,
			       total_no_of_primes_for_A,
			       M, Q_prime, candidates, tc,
			       start_index_FB_for_A, FREL, LPNEW);
    }
    else
      t = 0;

    total_full_relations += t;
    percentage =
      (long)((100.0 * total_full_relations) / (size_of_FB + 70));
    useless_candidates =
      (long)((100.0 * (total_candidates_number - total_full_relations))
	     / (total_candidates_number ? total_candidates_number : 1));

    if (percentage >= sort_interval)
    {
      if (DEBUGLEVEL >= 3)
      {
        if (DEBUGLEVEL >= 4)
          fprintferr("\nMPQS: passing the %ld%% mark, time = %ld ms\n",
                     sort_interval, timer2());
        else
          fprintferr("\nMPQS: passing the %ld%% mark\n", sort_interval);
        flusherr();
      }
      mpqs_close_file(LPNEW);
      mpqs_sort_lp_file(LPNEW_str);

      LPREL = mpqs_open_file(LPREL_str, "r");
      LPNEW = mpqs_open_file(LPNEW_str, "r");
      tp = mpqs_mergesort_lp_file(LPREL, LPNEW);
      /* opens LPTMP */
      mpqs_close_file(LPREL);
      mpqs_close_file(LPNEW);

      if (unlink(LPREL_str))
	err(warner, "can\'t remove file %s", LPREL_str);
      if (rename(LPTMP_str, LPREL_str))
	err(talker, "can\'t rename file %s to %s", LPTMP_str, LPREL_str);
      LPNEW = mpqs_open_file(LPNEW_str, "w");

      total_full_relations += tp;
      total_partial_relations += tp;
      percentage =
	(long)((100.0 * total_full_relations) / (size_of_FB + 70));
      vain_iterations =
	(long)((100.0 * (iterations - good_iterations))
	       / iterations);

      if (DEBUGLEVEL >= 4)
      {
        fprintferr("MPQS: found %ld%% of the required relations\n",
                   percentage);
        if (DEBUGLEVEL >= 5)
        {
          /* total_full_relations are always plural */
	  fprintferr("MPQS: found %ld full relations\n",
		     total_full_relations);
	  fprintferr("MPQS:   (%ld of these from partial relations)\n",
		     total_partial_relations);
	  fprintferr("MPQS: %ld%% useless candidates\n",
		     useless_candidates);
	  fprintferr("MPQS: %ld%% of the iterations yielded no candidates\n",
		     vain_iterations);
	}
      }

      if (percentage >= 90 && decimal_digits > 50)
	while (sort_interval <= percentage) sort_interval += 5;
      else
	while (sort_interval <= percentage) sort_interval += 10;

      if (percentage >= 100)
      {
	long total_relations;
	/* build full relations out of large prime relations */
	LPREL = mpqs_open_file(LPREL_str, "r");
	mpqs_combine_large_primes(
          LPREL, FREL, size_of_FB, kN
#ifdef MPQS_DEBUG
          , FB
#endif
        );
	mpqs_close_file(FREL);
	mpqs_close_file(LPREL);

	/* solve the system over F_2 */
	total_relations = mpqs_sort_lp_file(FREL_str);
	if (total_relations > size_of_FB)
	{
	  GEN fact;
	  if (DEBUGLEVEL >= 4)
	    fprintferr("\nMPQS: starting Gauss over F_2\n");
	  fact = mpqs_solve_linear_system(kN, N,
					  total_relations,
					  FB, size_of_FB);

	  /* if solution found, clear and return */
	  if (fact)
	  {
	    /* clean up */
	    CLEANUP();
	    mpqs_close_file(LPNEW);
	    if (unlink(FREL_str))
	      err(warner, "can\'t remove file %s", FREL_str);
	    if (unlink(LPREL_str))
	      err(warner, "can\'t remove file %s", LPREL_str);
	    if (unlink(LPNEW_str))
	      err(warner, "can\'t remove file %s", LPNEW_str);
	    if (DEBUGLEVEL >= 4)
            {
              fprintferr("\nMPQS: time in combining/Gauss/gcds = %ld ms\n",
                         timer2());
	      if (typ(fact) == t_INT)
		fprintferr("MPQS: found factor = %Z\n",
			   (long)fact);
	      else
	      {
		long j, nf=(lg(fact)-1)/3;
		if (nf==2)
		  fprintferr("MPQS: found factors = %Z\n\tand %Z\n",
			     /* (long) anyway */ fact[4], fact[1]);
		else
		{
		  fprintferr("MPQS: found %ld factors =\n", nf);
		  for (j=nf; j; j--)
		    fprintferr("\t%Z%s\n",
			       /* (long) anyway */ fact[3*j-2],
			       (j>1? "," : ""));
		}
	      }
            }
	    if (mpqs_diffptr == diffptr) mpqs_diffptr = NULL;
	    all_clean = 1;
	    /* nuisance: fact may not be safe for a gcopy(), and thus
	       gerepileupto(av, gcopy(fact)) will usually segfault on
	       one of the NULL markers.  However, it is already a nice
	       connected object, and it resides already the top of the
	       stack, so... :^) --GN */
	    return gerepileupto(av, fact);
	  }
	  else
	  {
	    if (DEBUGLEVEL >= 4)
	    {
              fprintferr("\nMPQS: time in combining/Gauss/gcds = %ld ms\n",
                         timer2());
	      fprintferr("MPQS: no factors found.\n");
	      if (percentage <= max_percentage)
		fprintferr("\nMPQS: restarting sieving ...\n");
	      else
		fprintferr("\nMPQS: giving up.\n");
	    }
	    if (percentage > max_percentage)
	    {
	      /* clean up */
	      CLEANUP();
	      mpqs_close_file(LPNEW);
	      if (unlink(FREL_str))
		err(warner, "can\'t remove file %s", FREL_str);
	      if (unlink(LPREL_str))
		err(warner, "can\'t remove file %s", LPREL_str);
	      if (unlink(LPNEW_str))
		err(warner, "can\'t remove file %s", LPNEW_str);
	      if (mpqs_diffptr == diffptr) mpqs_diffptr = NULL;
	      all_clean = 1;
	      avma=av; return NULL;
	    }
	    FREL = mpqs_open_file(FREL_str, "a");
	    LPREL = mpqs_open_file(LPREL_str, "a");
	  }
	}
        else
        {
          if (DEBUGLEVEL >= 4)
	  {
            fprintferr("\nMPQS: time in combining = %ld ms\n",
                       timer2());
            fprintferr("MPQS: too few independent relations found.\n");
	    if (percentage <= max_percentage)
	      fprintferr("\nMPQS: continuing sieving...\n");
	    else
	      fprintferr("\nMPQS: giving up.\n");
	  }
	  if (percentage > max_percentage)
	  {
	    /* clean up */
	    CLEANUP();
	    mpqs_close_file(LPNEW);
	    if (unlink(FREL_str))
	      err(warner, "can\'t remove file %s", FREL_str);
	    if (unlink(LPREL_str))
	      err(warner, "can\'t remove file %s", LPREL_str);
	    if (unlink(LPNEW_str))
	      err(warner, "can\'t remove file %s", LPNEW_str);
	    if (mpqs_diffptr == diffptr) mpqs_diffptr = NULL;
	    all_clean = 1;
	    avma=av; return NULL;
	  }
          FREL = mpqs_open_file(FREL_str, "a");
          LPREL = mpqs_open_file(LPREL_str, "a");
        }
      }
    }
  }
  return NULL;  /* not reached */
}
